import React from "react";
import { KeyCodes } from "@microsoft/fast-web-utilities";
import Foundation, {
    FoundationProps,
    HandledProps,
} from "@microsoft/fast-components-foundation-react";
import { get } from "lodash-es";
import { canUseDOM } from "exenv-es6";
import {
    NavigationDataType,
    NavigationHandledProps,
    NavigationProps,
    NavigationState,
    NavigationUnhandledProps,
    TreeNavigation,
} from "./navigation.props";
import { getNavigationFromData } from "./navigation.utilities";

export default class Navigation extends Foundation<
    NavigationHandledProps,
    NavigationUnhandledProps,
    NavigationState
> {
    public static displayName: string = "Navigation";

    public static getDerivedStateFromProps(
        props: NavigationProps,
        state: NavigationState
    ): Partial<NavigationState> {
        const updatedState: Partial<NavigationState> = {};
        const updatedNavigation: TreeNavigation = getNavigationFromData(
            props.data,
            props.schema,
            props.childOptions
        );

        if (props.dataLocation !== state.activeItem && props.dataLocation !== undefined) {
            updatedState.activeItem = props.dataLocation;

            // openItems is cast as any due to missing type on string array
            if (
                typeof props.dataLocation === "string" &&
                !(state.openItems as any).includes(props.dataLocation)
            ) {
                updatedState.openItems = state.openItems;
                updatedState.openItems.push(props.dataLocation);
            }
        }

        if (updatedNavigation !== state.navigation) {
            updatedState.navigation = updatedNavigation;
        }

        return updatedState;
    }

    protected handledProps: HandledProps<NavigationHandledProps> = {
        schema: void 0,
        data: void 0,
        childOptions: void 0,
        onLocationUpdate: void 0,
        dataLocation: void 0,
        managedClasses: void 0,
    };

    private rootElement: React.RefObject<HTMLDivElement>;

    constructor(props: NavigationProps) {
        super(props);

        this.state = {
            navigation: getNavigationFromData(
                this.props.data,
                this.props.schema,
                this.props.childOptions
            ),
            openItems:
                this.props.dataLocation !== undefined ? [this.props.dataLocation] : [],
            activeItem:
                this.props.dataLocation !== undefined ? this.props.dataLocation : null,
        };

        this.rootElement = React.createRef();
    }

    public render(): React.ReactNode {
        return (
            <div
                {...this.unhandledProps()}
                ref={this.rootElement}
                role={"tree"}
                className={this.props.managedClasses.navigation}
            >
                {this.renderTreeItem(this.state.navigation, 1, 1, 1, 0)}
            </div>
        );
    }

    /**
     * Renders the tree item containing element
     */
    private renderTreeItemContainer(
        navigation: TreeNavigation[],
        level: number
    ): React.ReactNode {
        return (
            <div className={this.props.managedClasses.navigation_itemList} role={"group"}>
                {this.renderTreeItems(navigation, level + 1)}
            </div>
        );
    }

    private renderTreeItem(
        navigation: TreeNavigation,
        level: number,
        navigationLength: number,
        positionInNavigation: number,
        index: number
    ): React.ReactNode {
        if (Array.isArray(navigation.items) && navigationLength > 0) {
            return (
                <div
                    className={this.props.managedClasses.navigation_item}
                    key={index}
                    role={"treeitem"}
                    aria-level={level}
                    aria-setsize={navigationLength}
                    aria-posinset={positionInNavigation}
                    aria-expanded={this.isExpanded(navigation.dataLocation)}
                    onClick={this.handleTreeItemClick(
                        navigation.dataLocation,
                        navigation.type
                    )}
                    onKeyUp={this.handleTreeItemKeyUp(
                        navigation.dataLocation,
                        navigation.type
                    )}
                >
                    <span
                        className={this.getTriggerClassNames(navigation.dataLocation)}
                        onClick={this.handleTreeItemClick(
                            navigation.dataLocation,
                            navigation.type
                        )}
                        onKeyUp={this.handleTreeItemKeyUp(
                            navigation.dataLocation,
                            navigation.type
                        )}
                        tabIndex={0}
                        data-location={navigation.dataLocation}
                    >
                        {navigation.text}
                    </span>
                    {this.renderTreeItemContainer(navigation.items, level)}
                </div>
            );
        }

        return (
            <div
                className={this.props.managedClasses.navigation_item}
                key={index}
                role={"none"}
            >
                <a
                    className={this.getLinkClassNames(navigation.dataLocation)}
                    role={"treeitem"}
                    data-location={navigation.dataLocation}
                    aria-level={level}
                    aria-setsize={navigationLength}
                    aria-posinset={positionInNavigation}
                    href={"#"}
                    onClick={this.handleTreeItemClick(
                        navigation.dataLocation,
                        navigation.type
                    )}
                    onKeyUp={this.handleTreeItemKeyUp(
                        navigation.dataLocation,
                        navigation.type
                    )}
                    tabIndex={0}
                >
                    {navigation.text}
                </a>
            </div>
        );
    }

    /**
     * Renders tree items
     */
    private renderTreeItems(
        navigation: TreeNavigation[],
        level: number
    ): React.ReactNode {
        return navigation.map((navigationItem: TreeNavigation, index: number) => {
            const navigationLength: number = navigation.length;
            const positionInNavigation: number = index + 1;

            return this.renderTreeItem(
                navigationItem,
                level,
                navigationLength,
                positionInNavigation,
                index
            );
        });
    }

    private findCurrentTreeItemIndex(nodes: HTMLElement[], dataLocation: string): number {
        return nodes.findIndex((node: HTMLElement) => {
            return node.dataset.location === dataLocation;
        });
    }

    private focusNextTreeItem(dataLocation: string): void {
        if (canUseDOM()) {
            const nodes: HTMLElement[] = this.getTreeItemNodes();
            const currentIndex: number = this.findCurrentTreeItemIndex(
                nodes,
                dataLocation
            );
            const nextIndex: number =
                currentIndex !== -1 && currentIndex !== nodes.length - 1
                    ? currentIndex + 1
                    : nodes.length - 1;
            nodes[nextIndex].focus();
        }
    }

    private focusPreviousTreeItem(dataLocation: string): void {
        if (canUseDOM()) {
            const nodes: HTMLElement[] = this.getTreeItemNodes();
            const currentIndex: number = this.findCurrentTreeItemIndex(
                nodes,
                dataLocation
            );
            const previousIndex: number =
                currentIndex !== -1 && currentIndex !== 0 ? currentIndex - 1 : 0;
            nodes[previousIndex].focus();
        }
    }

    private focusFirstTreeItem(): void {
        if (canUseDOM()) {
            const nodes: HTMLElement[] = this.getTreeItemNodes();

            nodes[0].focus();
        }
    }

    private focusLastTreeItem(): void {
        if (canUseDOM()) {
            const nodes: HTMLElement[] = this.getTreeItemNodes();

            nodes[nodes.length - 1].focus();
        }
    }

    private focusAndOpenTreeItems(dataLocation: string, type: NavigationDataType): void {
        if (canUseDOM()) {
            const nodes: HTMLElement[] = this.getTreeItemNodes();
            const currentIndex: number = this.findCurrentTreeItemIndex(
                nodes,
                dataLocation
            );
            const ariaExpanded: string = get(
                nodes[currentIndex],
                'parentElement.attributes["aria-expanded"].value'
            );

            if (ariaExpanded === "true" && nodes[currentIndex + 1]) {
                nodes[currentIndex + 1].focus();
            } else if (ariaExpanded === "false") {
                this.toggleItems(dataLocation, type);
            }
        }
    }

    private focusAndCloseTreeItems(dataLocation: string, type: NavigationDataType): void {
        if (canUseDOM()) {
            const nodes: HTMLElement[] = this.getTreeItemNodes();
            const currentIndex: number = this.findCurrentTreeItemIndex(
                nodes,
                dataLocation
            );
            const ariaExpanded: string = get(
                nodes[currentIndex],
                'parentElement.attributes["aria-expanded"].value'
            );

            if (
                ariaExpanded === "false" &&
                nodes[currentIndex - 1] &&
                (nodes[currentIndex - 1] === nodes[currentIndex].parentElement ||
                    nodes[currentIndex - 1].parentElement ===
                        get(
                            nodes[currentIndex],
                            "parentElement.parentElement.parentElement"
                        ))
            ) {
                nodes[currentIndex - 1].focus();
            } else if (ariaExpanded === "true") {
                this.toggleItems(dataLocation, type);
            } else if (ariaExpanded === undefined && nodes[currentIndex - 1]) {
                nodes[currentIndex - 1].focus();
            }
        }
    }

    private getLinkClassNames(dataLocation: string): string {
        let classes: string = get(this.props, "managedClasses.navigation_itemLink", "");

        if (dataLocation === this.state.activeItem) {
            classes = `${classes} ${get(
                this.props,
                "managedClasses.navigation_itemLink__active",
                ""
            )}`;
        }

        return classes;
    }

    private getTriggerClassNames(dataLocation: string): string {
        let classes: string = this.props.managedClasses.navigation_itemExpandListTrigger;

        if (dataLocation === this.state.activeItem) {
            classes = `${classes} ${get(
                this.props,
                "managedClasses.navigation_itemExpandListTrigger__active",
                ""
            )}`;
        }

        return classes;
    }

    private getTreeItemNodes(): HTMLElement[] {
        const nodes: HTMLElement[] = Array.from(
            this.rootElement.current.querySelectorAll(
                "a[role='treeitem'], div[role='treeitem'] > span"
            )
        );
        return nodes.filter((node: HTMLElement) => node.offsetParent !== null);
    }

    /**
     * Handles key up on a tree item
     */
    private handleTreeItemKeyUp = (
        dataLocation: string,
        type: NavigationDataType
    ): ((e: React.KeyboardEvent<HTMLDivElement | HTMLAnchorElement>) => void) => {
        return (e: React.KeyboardEvent<HTMLDivElement | HTMLAnchorElement>): void => {
            e.preventDefault();

            if (e.target === e.currentTarget) {
                switch (e.keyCode) {
                    case KeyCodes.enter:
                    case KeyCodes.space:
                        if (e.target === e.currentTarget) {
                            this.toggleItems(dataLocation, type);
                        }
                        break;
                    case KeyCodes.arrowDown:
                        this.focusNextTreeItem(dataLocation);
                        break;
                    case KeyCodes.arrowUp:
                        this.focusPreviousTreeItem(dataLocation);
                        break;
                    case KeyCodes.arrowRight:
                        this.focusAndOpenTreeItems(dataLocation, type);
                        break;
                    case KeyCodes.arrowLeft:
                        this.focusAndCloseTreeItems(dataLocation, type);
                        break;
                    case KeyCodes.home:
                        this.focusFirstTreeItem();
                        break;
                    case KeyCodes.end:
                        this.focusLastTreeItem();
                        break;
                    default:
                        break;
                }
            }
        };
    };

    /**
     * Handles clicking on a tree item
     */
    private handleTreeItemClick = (
        dataLocation: string,
        type: NavigationDataType
    ): ((e: React.MouseEvent<HTMLAnchorElement | HTMLDivElement>) => void) => {
        return (e: React.MouseEvent<HTMLAnchorElement>): void => {
            e.preventDefault();

            if (e.target === e.currentTarget) {
                this.toggleItems(dataLocation, type);
            }
        };
    };

    /**
     * Toggles the items by adding/removing them from the openItems array
     */
    private toggleItems(dataLocation: string, type: NavigationDataType): void {
        const isExpanded: boolean = this.isExpanded(dataLocation);
        const updatedState: Partial<NavigationState> = {};

        if (!isExpanded) {
            updatedState.openItems = this.state.openItems.concat(dataLocation);
        } else {
            updatedState.openItems = this.state.openItems.filter((openItem: string) => {
                return openItem.slice(0, dataLocation.length) !== dataLocation;
            });
        }

        if (this.props.dataLocation === undefined) {
            updatedState.activeItem = dataLocation;
        }

        this.setState(updatedState as NavigationState);

        if (
            typeof this.props.onLocationUpdate === "function" &&
            type !== NavigationDataType.children
        ) {
            this.props.onLocationUpdate(dataLocation);
        }
    }

    /**
     * Determines if the tree item should be expanded
     */
    private isExpanded(dataLocation: string): boolean {
        if (
            this.state.openItems.find(
                (openItem: string) =>
                    openItem.slice(0, dataLocation.length) === dataLocation
            ) !== undefined
        ) {
            return true;
        }

        return false;
    }
}
